package com.ho.memory.check.test1;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;


/**
 * This class is used to check / reserve the available memory space in VM for the execution of Java code in order to avoid the "out of memory" errors.
 * 
 * The "main" method simulates multiple threads call load a file (size of bytes) into memory Java (JVM).
 * 
 * 
 * @author Huseyin OZVEREN
 *
 */
public class MemoryChecker {
	// -------------------------------------- PRIVATE ATTRIBUTES
	// Number of tests of memory allocation for treatment before returning an OutOfMemory error
	private static int NB_OF_RETRY = 5;
	
	// If there is enough memory, the thread will be put on hold for the specified delay (between 2 retries).
	private static int WAITING_DELAY_MS = 500;
	
	// ------------------------------------------ PUBLIC METHODS
	/**
	 * Method to check if there is enough memory to process the given number of bytes.
	 * 
	 * If there is enough memory, the thread will be put on hold for the specified delay.
	 * 
	 * Example of use:
	 * Standby current thread until there will be enough memory space to handle the business
	 * 			MemoryChecker.checkMemoryAndIfNecessaryPutThreadOnHold(myDoc.getContentSize(), "myThread123");
	 * 
	 * @param memorySize
	 * @param threadName
	 */
	public static void checkMemoryAndIfNecessaryPutThreadOnHold(long memorySize, String threadName){
		{
			OutOfMemoryError ooMe = null;
			int nbLoop = NB_OF_RETRY;
			
			while(nbLoop>0){
/*
				try{
					byte[] tmp = new byte[(int) memorySize];

					System.out.println(threadName + " : there is enough memory for the current thread :" + memorySize/1024 + "ko");

					// At this stage, the amount of memory required is available
					break;
				}catch(OutOfMemoryError e){
					try{
						System.out.println(threadName + " : " + memorySize/1024 + "ko : Not enough memory for the current thread, try again in " + WAITING_DELAY_MS + " ms");
						
						// Print the memory status in console
						logMemoryStatusInConsole(threadName);
						
						// Sleep the current thread
						Thread.sleep(WAITING_DELAY_MS);
						
					}catch(Throwable e1){
					}//end-try
					ooMe = e;
					nbLoop--;
				}//end-try
					
*/
				// -------------- NEW VERSION
//				try{
					long freeNow = Runtime.getRuntime().freeMemory(); //
					long totalNow = Runtime.getRuntime().totalMemory(); // parameter Xms
					long maxPossible = Runtime.getRuntime().maxMemory(); // parameter Xmx
					long usedNow = totalNow - freeNow;
					long freeRealPossible = maxPossible - usedNow;
					
					if(freeRealPossible >= memorySize){
						// At this stage, the amount of memory required is available
						break;
						
					}else{
						System.out.println(threadName + " : " + memorySize/1024 + "ko : Not enough memory for the current thread, try again in " + WAITING_DELAY_MS + " ms");
						try{
							// Sleep the current thread
							Thread.sleep(WAITING_DELAY_MS);
						}catch(Throwable e1){
						}//end-try
						nbLoop--;
					}
/*				}catch(OutOfMemoryError e){
					try{
						System.out.println(threadName + " : " + memorySize/1024 + "ko : Not enough memory for the current thread, try again in " + WAITING_DELAY_MS + " ms");

						// Print the memory status in console
						logMemoryStatusInConsole(threadName);

						// Sleep the current thread
						Thread.sleep(WAITING_DELAY_MS);
					}catch(Throwable e1){
					}//end-try
					ooMe = e;
					nbLoop--;
				} //end-try
*/
			}//end-while
			
			if(0 == nbLoop){
				System.out.println(threadName + " : " + memorySize/1024 + "ko : Not enough memory for the current thread despite testing, so the propagation of OutOfMemoryError will be done.");
				// Print the memory status in console
				logMemoryStatusInConsole(threadName);
				throw (ooMe != null ? ooMe : new OutOfMemoryError());
			}//end-if
		}//end-block

		/*
		// Print the memory status in console before the garbage collector
		System.out.println(threadName + " : " + "Print the memory status in console before the garbage collector");
		logMemoryStatusInConsole(threadName);

		//Runs the garbage collector.
		Runtime.getRuntime().gc();
		
		// Print the memory status in console after the garbage collector
		System.out.println(threadName + " : " + "Print the memory status in console after the garbage collector");
		logMemoryStatusInConsole(threadName);
		*/
	}
	

	// ----------------------------------------- PRIVATE METHODS
	/**
	 * Print the memory status in console
	 * @param threadName
	 */
	private static void logMemoryStatusInConsole(String threadName){
		/*
		 	Used_now = Total_now - Free_now
			Free_real = Max - Used_now
			Max => parameter JVM '-Xmx' (ex: -Xmx1G)
			Total => parameter JVM '-Xms' (ex: -Xms1G)
			The total memory is reserved at the application server starting time.
			Blocking condition to execute the java main:  maxMemory >= totalMemory
		 */

		Runtime rt = Runtime.getRuntime();
		
		// Returns the total amount of memory in the Java virtual machine.
		// parameter Xms ; (ex: -Xms1G)
		long totalNow = rt.totalMemory(); 

		// Returns the maximum amount of memory that the Java virtual machine will attempt to use
		// parameter Xmx ; (ex: -Xmx1G)
		long maxPossible = rt.maxMemory();

		// Returns the amount of free memory in the Java Virtual Machine.
		long freeNow = rt.freeMemory();
		
		long usedNow = totalNow - freeNow;
		long freeRealPossible = maxPossible - usedNow;
		
		//
		System.out.println(threadName + " : " + "memorybroker : free "+ freeRealPossible/1024 + "ko : (total/max) " + totalNow/1024 + "ko/" + maxPossible/1024+"ko");
	}


	
	// ---------------------------------------------------- MAIN
	public static void main(String[] args) {
		
		// Number of threads
		int nbThreads = 2;

		// Memory needed for each thread
		final int memorySizeNeeded = 300000*1024;
		//final int memorySizeNeeded = 300*1024;

		// Creation of threads
		ExecutorService exec = Executors.newFixedThreadPool(nbThreads);
		for (int i = 0; i < nbThreads; i++) {
			final int valueI = i;
			
			Runnable runnable = new Runnable(){
				private String threadName = null;
				public void run(){
					this.threadName = "thread_"+valueI;
					
					try{
						System.out.println(threadName + ":-- START --");

						// Memory status before the memory checking.
						System.out.println(threadName + " : " + "Memory status bssefore the memory checking");
						logMemoryStatusInConsole(threadName);

						// Memory checking AND bo the business
						{
							// Memory checking
							checkMemoryAndIfNecessaryPutThreadOnHold(memorySizeNeeded, threadName);
							// Business
							byte[] tab = new byte[memorySizeNeeded];
						}

						logMemoryStatusInConsole(threadName);
						System.out.println(threadName + ":-- END --");
						
					}catch(Throwable th){
						System.out.println(threadName + ":-- ERROR --" + th);
					}//end-try
				}//end-run
			};
			//end-Runnable
			exec.submit(runnable);//end-submit
		}//end-for
		
	}
	
}
